﻿using System;
using System.Net;
using System.Threading;

namespace Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Adapters
{
    /// <summary>
    ///     This
    ///     <see cref="IHttpListener" /> implementation wraps around a
    ///     <see cref="HttpListener" /> instance.
    /// </summary>
    internal sealed class HttpListenerAdapter : WebDavDisposableBase, IHttpListener
    {
        /// <summary>
        ///     Initializes a new instance of the <see cref="HttpListenerAdapter" /> class.
        /// </summary>
        internal HttpListenerAdapter()
        {
            AdaptedInstance = new HttpListener
            {
                AuthenticationSchemes = AuthenticationSchemes.IntegratedWindowsAuthentication,
                UnsafeConnectionNtlmAuthentication = false
            };
        }

        /// <summary>
        ///     Waits for a request to come in to the web server and returns a
        ///     <see cref="IHttpListenerContext" /> adapter around it.
        /// </summary>
        /// <param name="abortEvent">
        ///     A
        ///     <see cref="EventWaitHandle" /> to use for aborting the wait. If this
        ///     event becomes set before a request comes in, this method will return
        ///     <c>null</c>.
        /// </param>
        /// <returns>
        ///     A
        ///     <see cref="IHttpListenerContext" /> adapter object for a request;
        ///     or
        ///     <c>null</c> if the wait for a request was aborted due to
        ///     <paramref name="abortEvent" /> being set.
        /// </returns>
        /// <exception cref="System.ArgumentNullException">abortEvent</exception>
        /// <exception cref="ArgumentNullException"><paramref name="abortEvent" /> is <c>null</c>.</exception>
        public IHttpListenerContext GetContext(EventWaitHandle abortEvent)
        {
            if (abortEvent == null)
                throw new ArgumentNullException(nameof(abortEvent));

            var ar = AdaptedInstance.BeginGetContext(null, null);
            var index = WaitHandle.WaitAny(new[] {abortEvent, ar.AsyncWaitHandle});
            if (index != 1) return null;
            var context = AdaptedInstance.EndGetContext(ar);
            return new HttpListenerContextAdapter(context);
        }

        /// <summary>
        ///     Gets the internal instance that was adapted for WebDAV#.
        /// </summary>
        /// <value>
        ///     The adapted instance.
        /// </value>
        public HttpListener AdaptedInstance { get; }

        /// <summary>
        ///     Gets the Uniform Resource Identifier (
        ///     <see cref="Uri" />) prefixes handled by the
        ///     adapted
        ///     <see cref="HttpListener" /> object.
        /// </summary>
        public HttpListenerPrefixCollection Prefixes => AdaptedInstance.Prefixes;

        /// <summary>
        ///     Allows the adapted <see cref="HttpListener" /> to receive incoming requests.
        /// </summary>
        public void Start()
        {
            AdaptedInstance.Start();
        }

        /// <summary>
        ///     Causes the adapted <see cref="HttpListener" /> to stop receiving incoming requests.
        /// </summary>
        public void Stop()
        {
            AdaptedInstance.Stop();
        }

        /// <summary>
        ///     Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing">
        ///     <c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only
        ///     unmanaged resources.
        /// </param>
        protected override void Dispose(bool disposing)
        {
            if (AdaptedInstance.IsListening)
                AdaptedInstance.Close();
        }
    }
}