﻿using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Text;

namespace Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Adapters
{
    /// <summary>
    ///     This
    ///     <see cref="IHttpListenerResponse" /> implementation wraps around a
    ///     <see cref="HttpListenerResponse" /> instance.
    /// </summary>
    internal sealed class HttpListenerResponseAdapter : IHttpListenerResponse
    {
        /// <summary>
        ///     Initializes a new instance of the <see cref="HttpListenerResponseAdapter" /> class.
        /// </summary>
        /// <param name="Response">The <see cref="HttpListenerResponse" /> to adapt for WebDAV#.</param>
        /// <exception cref="System.ArgumentNullException">Response</exception>
        /// <exception cref="ArgumentNullException"><paramref name="Response" /> is <c>null</c>.</exception>
        public HttpListenerResponseAdapter(HttpListenerResponse Response)
        {
            if (Response == null)
                throw new ArgumentNullException(nameof(Response));

            AdaptedInstance = Response;
        }

        /// <summary>
        ///     Gets the internal instance that was adapted for WebDAV#.
        /// </summary>
        /// <value>
        ///     The adapted instance.
        /// </value>
        public HttpListenerResponse AdaptedInstance { get; }

        /// <summary>
        ///     Gets or sets the HTTP status code to be returned to the client.
        /// </summary>
        public int StatusCode
        {
            get { return AdaptedInstance.StatusCode; }

            set { AdaptedInstance.StatusCode = value; }
        }

        /// <summary>
        ///     Gets or sets a text description of the HTTP <see cref="StatusCode">status code</see> returned to the client.
        /// </summary>
        public string StatusDescription
        {
            get { return AdaptedInstance.StatusDescription; }

            set { AdaptedInstance.StatusDescription = value ?? string.Empty; }
        }

        /// <summary>
        ///     Gets a <see cref="Stream" /> object to which a response can be written.
        /// </summary>
        public Stream OutputStream => AdaptedInstance.OutputStream;

        /// <summary>
        ///     Gets or sets the <see cref="Encoding" /> for this response's <see cref="OutputStream" />.
        /// </summary>
        public Encoding ContentEncoding
        {
            get { return AdaptedInstance.ContentEncoding; }

            set { AdaptedInstance.ContentEncoding = value; }
        }

        /// <summary>
        ///     Gets or sets the number of bytes in the body data included in the response.
        /// </summary>
        public long ContentLength64
        {
            get { return AdaptedInstance.ContentLength64; }

            set { AdaptedInstance.ContentLength64 = value; }
        }

        /// <summary>
        ///     Sends the response to the client and releases the resources held by the adapted
        ///     <see cref="HttpListenerResponse" /> instance.
        /// </summary>
        public void Close()
        {
            try
            {
                AdaptedInstance.Close();
            }
            catch (Exception ex)
            {
                Debug.Print(ex.ToString());
            }
        }

        /// <summary>
        ///     Appends a value to the specified HTTP header to be sent with the response.
        /// </summary>
        /// <param name="name">The name of the HTTP header to append the <paramref name="value" /> to.</param>
        /// <param name="value">The value to append to the <paramref name="name" /> header.</param>
        public void AppendHeader(string name, string value)
        {
            AdaptedInstance.AppendHeader(name, value);
        }
    }
}