using System;
using Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Exceptions;

namespace Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Stores.BaseClasses
{
    /// <summary>
    ///     This is the base class for <see cref="IWebDavStoreItem" /> implementations.
    /// </summary>
    public class WebDavStoreItemBase : IWebDavStoreItem
    {
        private string _path;

        /// <summary>
        ///     Initializes a new instance of the <see cref="WebDavStoreItemBase" /> class.
        /// </summary>
        /// <param name="path">The name of this <see cref="IWebDavStoreItem" /></param>
        /// <exception cref="System.ArgumentNullException">name</exception>
        /// <exception cref="ArgumentNullException"><paramref name="path" /> is <c>null</c>.</exception>
        protected WebDavStoreItemBase(string path)
        {
            if (string.IsNullOrWhiteSpace(path))
                throw new ArgumentNullException(nameof(path));

            _path = path;
        }

        /// <summary>
        ///     Called before the name of this <see cref="IWebDavStoreItem" /> is changing.
        /// </summary>
        /// <param name="oldName">The old name of this <see cref="IWebDavStoreItem" />.</param>
        /// <param name="newName">The new name of this <see cref="IWebDavStoreItem" />.</param>
        /// <returns>
        ///     <c>true</c> if the name change is allowed;
        ///     otherwise,
        ///     <c>false</c>.
        /// </returns>
        protected virtual bool OnNameChanging(string oldName, string newName)
        {
            return true;
        }

        /// <summary>
        ///     Called after the name of this <see cref="IWebDavStoreItem" /> has changed.
        /// </summary>
        /// <param name="oldName">The old name of this <see cref="IWebDavStoreItem" />.</param>
        /// <param name="newName">The new name of this <see cref="IWebDavStoreItem" />.</param>
        protected virtual void OnNameChanged(string oldName, string newName)
        {
        }

        #region IWebDAVStoreItem Members

        /// <summary>
        ///     Gets or sets the name of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        /// <exception cref="WebDavForbiddenException"></exception>
        public string Name
        {
            get { return _path; }

            set
            {
                var fixedName = (value ?? string.Empty).Trim();
                if (fixedName == _path) return;
                if (!OnNameChanging(_path, fixedName))
                    throw new WebDavForbiddenException();
                var oldName = _path;
                _path = fixedName;
                OnNameChanged(oldName, _path);
            }
        }

        /// <summary>
        ///     Gets the creation date of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public virtual DateTime CreationDate => DateTime.Now;

        /// <summary>
        ///     Gets the modification date of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public virtual DateTime ModificationDate => DateTime.Now;

        /// <summary>
        ///     Gets the path to this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public virtual string ItemPath => string.Empty;

        /// <summary>
        ///     Gets if this <see cref="IWebDavStoreItem" /> is a collection.
        /// </summary>
        public bool IsCollection => true;

        /// <summary>
        ///     Gets the hidden state of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public int Hidden => 0;

        #endregion
    }
}