using System;
using System.IO;
using System.Security.Principal;
using System.Threading;
using Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Stores.BaseClasses;

namespace Mtp2Dav.WebDAVSharp.Server._1d2086a502937936ebc6bfe19cfa15d855be1c31.Stores.DiskStore
{
    /// <summary>
    ///     This class implements a disk-based
    ///     <see cref="IWebDavStoreItem" /> which can be either
    ///     a folder on disk (
    ///     <see cref="WebDavDiskStoreCollection" />) or a file on disk
    ///     (
    ///     <see cref="WebDavDiskStoreDocument" />).
    /// </summary>
    public class WebDavDiskStoreItem : WebDavStoreItemBase
    {
        private readonly string _path;

        /// <summary>
        ///     Gets the Identity of the person logged on via HTTP Request.
        /// </summary>
        protected readonly WindowsIdentity Identity;

        /// <summary>
        ///     Initializes a new instance of the <see cref="WebDavDiskStoreItem" /> class.
        /// </summary>
        /// <see cref="WebDavDiskStoreCollection" />
        /// that contains this
        /// <see cref="WebDavDiskStoreItem" />
        /// ;
        /// or
        /// <c>null</c>
        /// if this is the root
        /// <see cref="WebDavDiskStoreCollection" />
        /// .
        /// <param name="path">The path that this <see cref="WebDavDiskStoreItem" /> maps to.</param>
        /// <exception cref="System.ArgumentNullException">path</exception>
        /// <exception cref="ArgumentNullException"><paramref name="path" /> is <c>null</c> or empty.</exception>
        protected WebDavDiskStoreItem(string path) : base(path)
        {
            if (string.IsNullOrWhiteSpace(path))
                throw new ArgumentNullException(nameof(path));

            _path = path;
            Identity = (WindowsIdentity) Thread.GetData(Thread.GetNamedDataSlot(WebDavServer.HttpUser));
        }

        /// <summary>
        ///     Gets the path to this <see cref="WebDavDiskStoreItem" />.
        /// </summary>
        public override string ItemPath => _path;

        #region IWebDAVStoreItem Members

        /// <summary>
        ///     Gets or sets the name of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Unable to rename item</exception>
        public new string Name
        {
            get { return Path.GetFileName(_path); }

            set { throw new InvalidOperationException("Unable to rename item"); }
        }

        /// <summary>
        ///     Gets if this <see cref="IWebDavStoreItem" /> is a collection.
        /// </summary>
        public new bool IsCollection
        {
            get
            {
                // get the file attributes for file or directory
                var attr = File.GetAttributes(_path);

                //detect whether its a directory or file
                return (attr & FileAttributes.Directory) == FileAttributes.Directory;
            }
        }

        /// <summary>
        ///     Gets the creation date of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public override DateTime CreationDate => File.GetCreationTime(_path);

        /// <summary>
        ///     Gets the modification date of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        public override DateTime ModificationDate => File.GetLastWriteTime(_path);

        /// <summary>
        ///     Gets the hidden state of this <see cref="IWebDavStoreItem" />.
        /// </summary>
        /// <remarks>
        ///     Source: <see href="http://stackoverflow.com/questions/3612035/c-sharp-check-if-a-directory-is-hidden" />
        /// </remarks>
        public new int Hidden
        {
            get
            {
                var dir = new DirectoryInfo(_path);
                return (dir.Attributes & FileAttributes.Hidden) != 0 ? 1 : 0;
            }
        }

        #endregion
    }
}