﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using System.Threading;
using WebDAVSharp.Server.Adapters;
using WebDAVSharp.Server.Exceptions;
using WebDAVSharp.Server.MethodHandlers;
using WebDAVSharp.Server.Stores;

namespace WebDAVSharp.Server.Utilities
{
    /// <summary>
    /// </summary>
    public class HttpReciever
    {
        private readonly string _name;

        private readonly IHttpListener _listener;
        private readonly Dictionary<string, IWebDavMethodHandler> _methodHandlers;
        private readonly WebDavServer _server;
        private readonly IWebDavStore _store;

        /// <summary>
        /// </summary>
        /// <param name="name"></param>
        /// <param name="listener"></param>
        /// <param name="handlers"></param>
        /// <param name="store"></param>
        /// <param name="server"></param>
        public HttpReciever(string name, ref IHttpListener listener, ref Dictionary<string, IWebDavMethodHandler> handlers, ref IWebDavStore store, WebDavServer server)
        {
            _listener = listener;
            _methodHandlers = handlers;
            _store = store;
            _server = server;
            _name = name;
        }

        private void ProcessRequest(object state)
        {
            var context = (IHttpListenerContext)state;
            var url = context.Request.Url.ToString().ToLower();

            if (url.EndsWith("desktop.ini") ||
                url.EndsWith("folder.gif") ||
                url.EndsWith("folder.jpg") ||
                url.EndsWith("thumbs.db"))
            {
                var ex = new WebDavNotFoundException();
                context.Response.StatusCode = ex.StatusCode;
                context.Response.StatusDescription = ex.StatusDescription;
                if (ex.Message != context.Response.StatusDescription)
                {
                    var buffer = Encoding.UTF8.GetBytes(ex.Message);
                    context.Response.ContentEncoding = Encoding.UTF8;
                    context.Response.ContentLength64 = buffer.Length;
                    context.Response.OutputStream.Write(buffer, 0, buffer.Length);
                    context.Response.OutputStream.Flush();
                }

                context.Response.Close();
                return;
            }

            var ident = _listener.GetIdentity(context);
            Thread.SetData(Thread.GetNamedDataSlot(WebDavServer.HttpUser), ident);
            _store.UserAuthenticated(ident);

            var method = context.Request.HttpMethod;

            try
            {
                try
                {
                    IWebDavMethodHandler methodHandler;

                    if (!_methodHandlers.TryGetValue(method, out methodHandler))
                        throw new WebDavMethodNotAllowedException(string.Format(CultureInfo.InvariantCulture, "%s ({0})", context.Request.HttpMethod));

                    context.Response.AppendHeader("DAV", "1,2,1#extend");

                    methodHandler.ProcessRequest(_server, context, _store);
                }
                catch (WebDavException)
                {
                    throw;
                }
                catch (UnauthorizedAccessException)
                {
                    throw new WebDavUnauthorizedException();
                }
                catch (FileNotFoundException ex)
                {
                    throw new WebDavNotFoundException(innerException: ex);
                }
                catch (DirectoryNotFoundException ex)
                {
                    throw new WebDavNotFoundException(innerException: ex);
                }
                catch (NotImplementedException ex)
                {
                    throw new WebDavNotImplementedException(innerException: ex);
                }
                catch (Exception ex)
                {
                    throw new WebDavInternalServerException(innerException: ex);
                }
            }
            catch (WebDavException ex)
            {
                if (ex.Message != "Not Found")
                {
                    Console.WriteLine(ex.Message + ex.StackTrace);
                    if (ex.InnerException != null)
                        Console.WriteLine(ex.InnerException.Message + ex.InnerException.StackTrace);
                }

                context.Response.StatusCode = ex.StatusCode;
                context.Response.StatusDescription = ex.StatusDescription;
                if (ex.Message != context.Response.StatusDescription)
                {
                    var buffer = Encoding.UTF8.GetBytes(ex.Message);
                    context.Response.ContentEncoding = Encoding.UTF8;
                    context.Response.ContentLength64 = buffer.Length;
                    context.Response.OutputStream.Write(buffer, 0, buffer.Length);
                    context.Response.OutputStream.Flush();
                }

                context.Response.Close();
            }
        }

        // ReSharper disable once InconsistentNaming
        internal bool _stop;

        /// <summary>
        /// </summary>
        public void Start()
        {
            _stop = false;
            _listener.AdaptedInstance.BeginGetContext(OnContext, this);
        }

        /// <summary>
        /// </summary>
        /// <returns></returns>
        public void Stop()
        {
            _stop = true;
        }

        private static void OnContext(IAsyncResult ar)
        {
            var server = (HttpReciever)ar.AsyncState;
            if (!server._stop)
                server._listener.AdaptedInstance.BeginGetContext(OnContext, server);
            IHttpListenerContext context = new HttpListenerContextAdapter(server._listener.AdaptedInstance.EndGetContext(ar));
            server.ProcessRequest(context);
        }
    }
}