/* Copyright (C) 2014-2017 Tal Aloni <tal.aloni.il@gmail.com>. All rights reserved.
 * 
 * You can redistribute this program and/or modify it under the terms of
 * the GNU Lesser Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version.
 */
using System;
using System.Collections.Generic;
using System.IO;
using Utilities;

namespace SMBLibrary
{
    public class NTFileSystemAdapter : INTFileStore
    {
        public NTStatus GetFileInformation(out FileInformation result, object handle, FileInformationClass informationClass)
        {
            FileHandle fileHandle = (FileHandle)handle;
            string path = fileHandle.Path;
            IFileSystemEntry entry;
            try
            {
                entry = m_fileSystem.GetEntry(path);
            }
            catch (Exception ex)
            {
                NTStatus status = ToNTStatus(ex);
                Log(Severity.Verbose, "GetFileInformation on '{0}' failed. {1}", path, status);
                result = null;
                return status;
            }

            if (entry == null)
            {
                result = null;
                return NTStatus.STATUS_NO_SUCH_FILE;
            }

            switch (informationClass)
            {
                case FileInformationClass.FileBasicInformation:
                {
                    FileBasicInformation information = new FileBasicInformation();
                    information.CreationTime = entry.CreationTime;
                    information.LastAccessTime = entry.LastAccessTime;
                    information.LastWriteTime = entry.LastWriteTime;
                    information.ChangeTime = entry.LastWriteTime;
                    information.FileAttributes = GetFileAttributes(entry);
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileStandardInformation:
                {
                    FileStandardInformation information = new FileStandardInformation();
                    information.AllocationSize = (long)GetAllocationSize(entry.Size);
                    information.EndOfFile = (long)entry.Size;
                    information.Directory = entry.IsDirectory;
                    information.DeletePending = fileHandle.DeleteOnClose;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileInternalInformation:
                {
                    FileInternalInformation information = new FileInternalInformation();
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileEaInformation:
                {
                    FileEaInformation information = new FileEaInformation();
                    information.EaSize = 0;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileAccessInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileNameInformation:
                {
                    FileNameInformation information = new FileNameInformation();
                    information.FileName = entry.Name;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FilePositionInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileFullEaInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileModeInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileAlignmentInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileAllInformation:
                {
                    FileAllInformation information = new FileAllInformation();
                    information.BasicInformation.CreationTime = entry.CreationTime;
                    information.BasicInformation.LastAccessTime = entry.LastAccessTime;
                    information.BasicInformation.LastWriteTime = entry.LastWriteTime;
                    information.BasicInformation.ChangeTime = entry.LastWriteTime;
                    information.BasicInformation.FileAttributes = GetFileAttributes(entry);
                    information.StandardInformation.AllocationSize = (long)GetAllocationSize(entry.Size);
                    information.StandardInformation.EndOfFile = (long)entry.Size;
                    information.StandardInformation.Directory = entry.IsDirectory;
                    information.StandardInformation.DeletePending = fileHandle.DeleteOnClose;
                    information.NameInformation.FileName = entry.Name;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileAlternateNameInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileStreamInformation:
                {
                    // This information class is used to enumerate the data streams of a file or a directory.
                    // A buffer of FileStreamInformation data elements is returned by the server.
                    FileStreamInformation information = new FileStreamInformation();
                    FileStreamEntry streamEntry = new FileStreamEntry();
                    streamEntry.StreamSize = (long)entry.Size;
                    streamEntry.StreamAllocationSize = (long)GetAllocationSize(entry.Size);
                    streamEntry.StreamName = "::$DATA";
                    information.Entries.Add(streamEntry);
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FilePipeInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FilePipeLocalInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FilePipeRemoteInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileCompressionInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                case FileInformationClass.FileNetworkOpenInformation:
                {
                    FileNetworkOpenInformation information = new FileNetworkOpenInformation();
                    information.CreationTime = entry.CreationTime;
                    information.LastAccessTime = entry.LastAccessTime;
                    information.LastWriteTime = entry.LastWriteTime;
                    information.ChangeTime = entry.LastWriteTime;
                    information.AllocationSize = (long)GetAllocationSize(entry.Size);
                    information.EndOfFile = (long)entry.Size;
                    information.FileAttributes = GetFileAttributes(entry);
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileInformationClass.FileAttributeTagInformation:
                {
                    result = null;
                    return NTStatus.STATUS_NOT_IMPLEMENTED;
                }
                default:
                    result = null;
                    return NTStatus.STATUS_INVALID_INFO_CLASS;
            }
        }

        public static FileAttributes GetFileAttributes(IFileSystemEntry entry)
        {
            FileAttributes attributes = 0;
            if (entry.IsHidden)
            {
                attributes |= FileAttributes.Hidden;
            }
            if (entry.IsReadonly)
            {
                attributes |= FileAttributes.ReadOnly;
            }
            if (entry.IsArchived)
            {
                attributes |= FileAttributes.Archive;
            }
            if (entry.IsDirectory)
            {
                attributes |= FileAttributes.Directory;
            }

            if (attributes == 0)
            {
                attributes = FileAttributes.Normal;
            }

            return attributes;
        }

        private const int BytesPerSector = 512;
        private const int ClusterSize = 4096;

        private IFileSystem m_fileSystem;

        public event EventHandler<LogEntry> LogEntryAdded;

        public NTFileSystemAdapter(IFileSystem fileSystem)
        {
            m_fileSystem = fileSystem;
        }

        public NTStatus CreateFile(out object handle, out FileStatus fileStatus, string path, AccessMask desiredAccess, FileAttributes fileAttributes, ShareAccess shareAccess, CreateDisposition createDisposition, CreateOptions createOptions, SecurityContext securityContext)
        {
            handle = null;
            fileStatus = FileStatus.FILE_DOES_NOT_EXIST;
            var createAccess = NTFileStoreHelper.ToCreateFileAccess(desiredAccess, createDisposition);
            var requestedWriteAccess = (createAccess & FileAccess.Write) > 0;

            var forceDirectory = (createOptions & CreateOptions.FILE_DIRECTORY_FILE) > 0;
            var forceFile = (createOptions & CreateOptions.FILE_NON_DIRECTORY_FILE) > 0;

            if (forceDirectory & (createDisposition != CreateDisposition.FILE_CREATE &&
                                  createDisposition != CreateDisposition.FILE_OPEN &&
                                  createDisposition != CreateDisposition.FILE_OPEN_IF &&
                                  createDisposition != CreateDisposition.FILE_SUPERSEDE))
            {
                return NTStatus.STATUS_INVALID_PARAMETER;
            }

            // Windows will try to access named streams (alternate data streams) regardless of the FILE_NAMED_STREAMS flag, we need to prevent this behaviour.
            if (path.Contains(":"))
            {
                // Windows Server 2003 will return STATUS_OBJECT_NAME_NOT_FOUND
                return NTStatus.STATUS_NO_SUCH_FILE;
            }

            IFileSystemEntry entry;
            try
            {
                entry = m_fileSystem.GetEntry(path);
            }
            catch (Exception ex)
            {
                var status = ToNTStatus(ex);
                Log(Severity.Verbose, "CreateFile: Error retrieving '{0}'. {1}.", path, status);
                return status;
            }

            if (createDisposition == CreateDisposition.FILE_OPEN)
            {
                if (entry == null)
                {
                    return NTStatus.STATUS_NO_SUCH_FILE;
                }

                fileStatus = FileStatus.FILE_EXISTS;
                if (entry.IsDirectory && forceFile)
                {
                    return NTStatus.STATUS_FILE_IS_A_DIRECTORY;
                }

                if (!entry.IsDirectory && forceDirectory)
                {
                    return NTStatus.STATUS_OBJECT_PATH_INVALID;
                }
            }
            else if (createDisposition == CreateDisposition.FILE_CREATE)
            {
                if (entry != null)
                {
                    // File already exists, fail the request
                    Log(Severity.Verbose, "CreateFile: File '{0}' already exists.", path);
                    fileStatus = FileStatus.FILE_EXISTS;
                    return NTStatus.STATUS_OBJECT_NAME_COLLISION;
                }

                if (!requestedWriteAccess)
                {
                    return NTStatus.STATUS_ACCESS_DENIED;
                }

                try
                {
                    if (forceDirectory)
                    {
                        Log(Severity.Information, "CreateFile: Creating directory '{0}'", path);
                        entry = m_fileSystem.CreateDirectory(path);
                    }
                    else
                    {
                        Log(Severity.Information, "CreateFile: Creating file '{0}'", path);
                        entry = m_fileSystem.CreateFile(path);
                    }
                }
                catch (Exception ex)
                {
                    var status = ToNTStatus(ex);
                    Log(Severity.Verbose, "CreateFile: Error creating '{0}'. {1}.", path, status);
                    return status;
                }
                fileStatus = FileStatus.FILE_CREATED;
            }
            else if (createDisposition == CreateDisposition.FILE_OPEN_IF ||
                     createDisposition == CreateDisposition.FILE_OVERWRITE ||
                     createDisposition == CreateDisposition.FILE_OVERWRITE_IF ||
                     createDisposition == CreateDisposition.FILE_SUPERSEDE)
            {
                if (entry == null)
                {
                    if (createDisposition == CreateDisposition.FILE_OVERWRITE)
                    {
                        return NTStatus.STATUS_OBJECT_PATH_NOT_FOUND;
                    }

                    if (!requestedWriteAccess)
                    {
                        return NTStatus.STATUS_ACCESS_DENIED;
                    }

                    try
                    {
                        if (forceDirectory)
                        {
                            Log(Severity.Information, "CreateFile: Creating directory '{0}'", path);
                            entry = m_fileSystem.CreateDirectory(path);
                        }
                        else
                        {
                            Log(Severity.Information, "CreateFile: Creating file '{0}'", path);
                            entry = m_fileSystem.CreateFile(path);
                        }
                    }
                    catch (Exception ex)
                    {
                        var status = ToNTStatus(ex);
                        Log(Severity.Verbose, "CreateFile: Error creating '{0}'. {1}.", path, status);
                        return status;
                    }
                    fileStatus = FileStatus.FILE_CREATED;
                }
                else
                {
                    fileStatus = FileStatus.FILE_EXISTS;
                    if (createDisposition == CreateDisposition.FILE_OPEN_IF)
                    {
                        if (entry.IsDirectory && forceFile)
                        {
                            return NTStatus.STATUS_FILE_IS_A_DIRECTORY;
                        }

                        if (!entry.IsDirectory && forceDirectory)
                        {
                            return NTStatus.STATUS_OBJECT_PATH_INVALID;
                        }
                    }
                    else
                    {
                        if (!requestedWriteAccess)
                        {
                            return NTStatus.STATUS_ACCESS_DENIED;
                        }

                        if (createDisposition == CreateDisposition.FILE_OVERWRITE ||
                            createDisposition == CreateDisposition.FILE_OVERWRITE_IF)
                        {
                            // Truncate the file
                            try
                            {
                                var temp = m_fileSystem.OpenFile(path, FileMode.Truncate, FileAccess.ReadWrite, FileShare.ReadWrite, FileOptions.None);
                                temp.Close();
                            }
                            catch (Exception ex)
                            {
                                var status = ToNTStatus(ex);
                                Log(Severity.Verbose, "CreateFile: Error truncating '{0}'. {1}.", path, status);
                                return status;
                            }
                            fileStatus = FileStatus.FILE_OVERWRITTEN;
                        }
                        else if (createDisposition == CreateDisposition.FILE_SUPERSEDE)
                        {
                            // Delete the old file
                            try
                            {
                                m_fileSystem.Delete(path);
                            }
                            catch (Exception ex)
                            {
                                var status = ToNTStatus(ex);
                                Log(Severity.Verbose, "CreateFile: Error deleting '{0}'. {1}.", path, status);
                                return status;
                            }

                            try
                            {
                                if (forceDirectory)
                                {
                                    Log(Severity.Information, "CreateFile: Creating directory '{0}'", path);
                                    entry = m_fileSystem.CreateDirectory(path);
                                }
                                else
                                {
                                    Log(Severity.Information, "CreateFile: Creating file '{0}'", path);
                                    entry = m_fileSystem.CreateFile(path);
                                }
                            }
                            catch (Exception ex)
                            {
                                var status = ToNTStatus(ex);
                                Log(Severity.Verbose, "CreateFile: Error creating '{0}'. {1}.", path, status);
                                return status;
                            }
                            fileStatus = FileStatus.FILE_SUPERSEDED;
                        }
                    }
                }
            }
            else
            {
                return NTStatus.STATUS_INVALID_PARAMETER;
            }

            var fileAccess = NTFileStoreHelper.ToFileAccess((FileAccessMask)desiredAccess);
            Stream stream;
            if (fileAccess == (FileAccess)0 || entry.IsDirectory)
            {
                stream = null;
            }
            else
            {
                // Note that SetFileInformationByHandle/FILE_DISPOSITION_INFO has no effect if the handle was opened with FILE_DELETE_ON_CLOSE.
                var openStatus = OpenFileStream(out stream, path, fileAccess, shareAccess, createOptions);
                if (openStatus != NTStatus.STATUS_SUCCESS)
                {
                    return openStatus;
                }
            }

            var deleteOnClose = (createOptions & CreateOptions.FILE_DELETE_ON_CLOSE) > 0;
            handle = new FileHandle(path, entry.IsDirectory, stream, deleteOnClose);
            if (fileStatus != FileStatus.FILE_CREATED &&
                fileStatus != FileStatus.FILE_OVERWRITTEN &&
                fileStatus != FileStatus.FILE_SUPERSEDED)
            {
                fileStatus = FileStatus.FILE_OPENED;
            }
            return NTStatus.STATUS_SUCCESS;
        }

        private NTStatus OpenFileStream(out Stream stream, string path, FileAccess fileAccess, ShareAccess shareAccess, CreateOptions openOptions)
        {
            stream = null;
            var fileShare = NTFileStoreHelper.ToFileShare(shareAccess);
            var fileOptions = ToFileOptions(openOptions);
            var fileShareString = fileShare.ToString().Replace(", ", "|");
            var fileOptionsString = ToFileOptionsString(fileOptions);
            try
            {
                stream = m_fileSystem.OpenFile(path, FileMode.Open, fileAccess, fileShare, fileOptions);
            }
            catch (Exception ex)
            {
                var status = ToNTStatus(ex);
                Log(Severity.Verbose, "OpenFile: Cannot open '{0}', Access={1}, Share={2}. NTStatus: {3}.", path, fileAccess, fileShareString, status);
                return status;
            }

            Log(Severity.Information, "OpenFileStream: Opened '{0}', Access={1}, Share={2}, FileOptions={3}", path, fileAccess, fileShareString, fileOptionsString);
            return NTStatus.STATUS_SUCCESS;
        }

        public NTStatus CloseFile(object handle)
        {
            var fileHandle = (FileHandle)handle;
            if (fileHandle.Stream != null)
            {
                Log(Severity.Verbose, "CloseFile: Closing '{0}'.", fileHandle.Path);
                fileHandle.Stream.Close();
            }

            // If the file / directory was created with FILE_DELETE_ON_CLOSE but was not opened (with FileOptions.DeleteOnClose), we should delete it now.
            if (fileHandle.Stream == null && fileHandle.DeleteOnClose)
            {
                try
                {
                    m_fileSystem.Delete(fileHandle.Path);
                    Log(Severity.Verbose, "CloseFile: Deleted '{0}'.", fileHandle.Path);
                }
                catch
                {
                    Log(Severity.Verbose, "CloseFile: Error deleting '{0}'.", fileHandle.Path);
                }
            }

            return NTStatus.STATUS_SUCCESS;
        }

        public NTStatus ReadFile(out byte[] data, object handle, long offset, int maxCount)
        {
            data = null;
            var fileHandle = (FileHandle)handle;
            var path = fileHandle.Path;
            var stream = fileHandle.Stream;
            if (stream == null || !stream.CanRead)
            {
                Log(Severity.Verbose, "ReadFile: Cannot read '{0}', Invalid Operation.", path);
                return NTStatus.STATUS_ACCESS_DENIED;
            }

            int bytesRead;
            try
            {
                stream.Seek(offset, SeekOrigin.Begin);
                data = new byte[maxCount];
                bytesRead = stream.Read(data, 0, maxCount);
            }
            catch (Exception ex)
            {
                var status = ToNTStatus(ex);
                Log(Severity.Verbose, "ReadFile: Cannot read '{0}'. {1}.", path, status);
                return status;
            }

            if (bytesRead < maxCount)
            {
                // EOF, we must trim the response data array
                data = ByteReader.ReadBytes(data, 0, bytesRead);
            }
            return NTStatus.STATUS_SUCCESS;
        }

        public NTStatus WriteFile(out int numberOfBytesWritten, object handle, long offset, byte[] data)
        {
            numberOfBytesWritten = 0;
            var fileHandle = (FileHandle)handle;
            var path = fileHandle.Path;
            var stream = fileHandle.Stream;
            if (stream == null || !stream.CanWrite)
            {
                Log(Severity.Verbose, "WriteFile: Cannot write '{0}'. Invalid Operation.", path);
                return NTStatus.STATUS_ACCESS_DENIED;
            }

            try
            {
                stream.Seek(offset, SeekOrigin.Begin);
                stream.Write(data, 0, data.Length);
            }
            catch (Exception ex)
            {
                var status = ToNTStatus(ex);
                Log(Severity.Verbose, "WriteFile: Cannot write '{0}'. {1}.", path, status);
                return status;
            }
            numberOfBytesWritten = data.Length;
            return NTStatus.STATUS_SUCCESS;
        }

        public NTStatus FlushFileBuffers(object handle)
        {
            var fileHandle = (FileHandle)handle;
            if (fileHandle.Stream != null)
            {
                fileHandle.Stream.Flush();
            }
            return NTStatus.STATUS_SUCCESS;
        }

        public NTStatus LockFile(object handle, long byteOffset, long length, bool exclusiveLock)
        {
            return NTStatus.STATUS_NOT_SUPPORTED;
        }

        public NTStatus UnlockFile(object handle, long byteOffset, long length)
        {
            return NTStatus.STATUS_NOT_SUPPORTED;
        }

        public NTStatus NotifyChange(out object ioRequest, object handle, NotifyChangeFilter completionFilter, bool watchTree, int outputBufferSize, OnNotifyChangeCompleted onNotifyChangeCompleted, object context)
        {
            ioRequest = null;
            return NTStatus.STATUS_NOT_SUPPORTED;
        }

        public NTStatus Cancel(object ioRequest)
        {
            return NTStatus.STATUS_NOT_SUPPORTED;
        }

        public NTStatus DeviceIOControl(object handle, uint ctlCode, byte[] input, out byte[] output, int maxOutputLength)
        {
            output = null;
            return NTStatus.STATUS_NOT_SUPPORTED;
        }

        public void Log(Severity severity, string message)
        {
            // To be thread-safe we must capture the delegate reference first
            var handler = LogEntryAdded;
            if (handler != null)
            {
                handler(this, new LogEntry(DateTime.Now, severity, "NT FileSystem Adapter", message));
            }
        }

        public void Log(Severity severity, string message, params object[] args)
        {
            Log(severity, String.Format(message, args));
        }

        /// <param name="exception">IFileSystem exception</param>
        private static NTStatus ToNTStatus(Exception exception)
        {
            if (exception is NotImplementedException) return NTStatus.STATUS_NOT_IMPLEMENTED;
            if (exception is NotSupportedException) return NTStatus.STATUS_NOT_SUPPORTED;
            if (exception is ArgumentException) return NTStatus.STATUS_OBJECT_PATH_SYNTAX_BAD;
            if (exception is DirectoryNotFoundException) return NTStatus.STATUS_OBJECT_PATH_NOT_FOUND;
            if (exception is FileNotFoundException) return NTStatus.STATUS_OBJECT_PATH_NOT_FOUND;
            if (exception is UnauthorizedAccessException)return NTStatus.STATUS_ACCESS_DENIED;

            if (exception is IOException ioException)
            {
                var errorCode = IOExceptionHelper.GetWin32ErrorCode(ioException);
                if (errorCode == (ushort)Win32Error.ERROR_SHARING_VIOLATION) return NTStatus.STATUS_SHARING_VIOLATION;
                if (errorCode == (ushort)Win32Error.ERROR_DISK_FULL) return NTStatus.STATUS_DISK_FULL;
                // If a user tries to rename folder1 to folder2 when folder2 already exists, Windows 7 will offer to merge folder1 into folder2.
                // In such case, Windows 7 will delete folder 1 and will expect STATUS_DIRECTORY_NOT_EMPTY if there are files to merge.
                if (errorCode == (ushort)Win32Error.ERROR_DIR_NOT_EMPTY) return NTStatus.STATUS_DIRECTORY_NOT_EMPTY;
                // According to [MS-FSCC], FileRenameInformation MUST return STATUS_OBJECT_NAME_COLLISION when the specified name already exists and ReplaceIfExists is zero.
                if (errorCode == (ushort)Win32Error.ERROR_ALREADY_EXISTS) return NTStatus.STATUS_OBJECT_NAME_COLLISION;
            }

            return NTStatus.STATUS_DATA_ERROR;
        }

        private static FileOptions ToFileOptions(CreateOptions createOptions)
        {
            const FileOptions FILE_FLAG_OPEN_REPARSE_POINT = (FileOptions)0x00200000;
            const FileOptions FILE_FLAG_NO_BUFFERING = (FileOptions)0x20000000;
            var result = FileOptions.None;
            if ((createOptions & CreateOptions.FILE_OPEN_REPARSE_POINT) > 0)
            {
                result |= FILE_FLAG_OPEN_REPARSE_POINT;
            }
            if ((createOptions & CreateOptions.FILE_NO_INTERMEDIATE_BUFFERING) > 0)
            {
                result |= FILE_FLAG_NO_BUFFERING;
            }
            if ((createOptions & CreateOptions.FILE_RANDOM_ACCESS) > 0)
            {
                result |= FileOptions.RandomAccess;
            }
            if ((createOptions & CreateOptions.FILE_SEQUENTIAL_ONLY) > 0)
            {
                result |= FileOptions.SequentialScan;
            }
            if ((createOptions & CreateOptions.FILE_WRITE_THROUGH) > 0)
            {
                result |= FileOptions.WriteThrough;
            }
            if ((createOptions & CreateOptions.FILE_DELETE_ON_CLOSE) > 0)
            {
                result |= FileOptions.DeleteOnClose;
            }

            return result;
        }

        private static string ToFileOptionsString(FileOptions options)
        {
            var result = String.Empty;
            const FileOptions FILE_FLAG_OPEN_REPARSE_POINT = (FileOptions)0x00200000;
            const FileOptions FILE_FLAG_NO_BUFFERING = (FileOptions)0x20000000;
            if ((options & FILE_FLAG_OPEN_REPARSE_POINT) > 0)
            {
                result += "ReparsePoint|";
                options &= ~FILE_FLAG_OPEN_REPARSE_POINT;
            }
            if ((options & FILE_FLAG_NO_BUFFERING) > 0)
            {
                result += "NoBuffering|";
                options &= ~FILE_FLAG_NO_BUFFERING;
            }

            if (result == String.Empty || options != FileOptions.None)
            {
                result += options.ToString().Replace(", ", "|");
            }
            result = result.TrimEnd(new char[] { '|' });
            return result;
        }

        /// <summary>
        /// Will return a virtual allocation size, assuming 4096 bytes per cluster
        /// </summary>
        public static ulong GetAllocationSize(ulong size)
        {
            return (ulong)Math.Ceiling((double)size / ClusterSize) * ClusterSize;
        }

        public static string GetShortName(string fileName)
        {
            var fileNameWithoutExt = System.IO.Path.GetFileNameWithoutExtension(fileName);
            var extension = System.IO.Path.GetExtension(fileName);
            if (fileNameWithoutExt.Length > 8 || extension.Length > 4)
            {
                if (fileNameWithoutExt.Length > 8)
                {
                    fileNameWithoutExt = fileNameWithoutExt.Substring(0, 8);
                }

                if (extension.Length > 4)
                {
                    extension = extension.Substring(0, 4);
                }

                return fileNameWithoutExt + extension;
            }
            return fileName;
        }

        /// <param name="fileName">Expression as described in [MS-FSA] 2.1.4.4</param>
        public NTStatus QueryDirectory(out List<QueryDirectoryFileInformation> result, object handle, string fileName, FileInformationClass informationClass)
        {
            result = null;
            var directoryHandle = (FileHandle)handle;
            if (!directoryHandle.IsDirectory)
            {
                return NTStatus.STATUS_INVALID_PARAMETER;
            }

            if (fileName == string.Empty)
            {
                return NTStatus.STATUS_INVALID_PARAMETER;
            }

            var path = directoryHandle.Path;
            var findExactName = !ContainsWildcardCharacters(fileName);

            List<IFileSystemEntry> entries;
            if (!findExactName)
            {
                try
                {
                    entries = m_fileSystem.ListEntriesInDirectory(path);
                } 
                catch (Exception ex)
                {
                    return ToNTStatus(ex);
                }

                entries = GetFiltered(entries, fileName);

                // Windows will return "." and ".." when enumerating directory files.
                // The SMB1 / SMB2 specifications mandate that when zero entries are found, the server SHOULD / MUST return STATUS_NO_SUCH_FILE.
                // For this reason, we MUST include the current directory and/or parent directory when enumerating a directory
                // in order to diffrentiate between a directory that does not exist and a directory with no entries.
                var currentDirectory = m_fileSystem.GetEntry(path).Clone();
                currentDirectory.Name = ".";
                var parentDirectory = m_fileSystem.GetEntry(FileSystem.GetParentDirectory(path)).Clone();
                parentDirectory.Name = "..";
                entries.Insert(0, parentDirectory);
                entries.Insert(0, currentDirectory);
            }
            else
            {
                path = FileSystem.GetDirectoryPath(path);
                var entry = m_fileSystem.GetEntry(path + fileName);
                if (entry == null)
                {
                    return NTStatus.STATUS_NO_SUCH_FILE;
                }
                entries = new List<IFileSystemEntry>();
                entries.Add(entry);
            }

            try
            {
                result = FromFileSystemEntries(entries, informationClass);
            }
            catch (UnsupportedInformationLevelException)
            {
                return NTStatus.STATUS_INVALID_INFO_CLASS;
            }
            return NTStatus.STATUS_SUCCESS;
        }

        /// <param name="expression">Expression as described in [MS-FSA] 2.1.4.4</param>
        private static List<IFileSystemEntry> GetFiltered(List<IFileSystemEntry> entries, string expression)
        {
            if (expression == "*")
            {
                return entries;
            }

            var result = new List<IFileSystemEntry>();
            foreach (var entry in entries)
            {
                if (IsFileNameInExpression(entry.Name, expression))
                {
                    result.Add(entry);
                }
            }
            return result;
        }

        private static bool ContainsWildcardCharacters(string expression)
        {
            return (expression.Contains("?") || expression.Contains("*") || expression.Contains("\"") || expression.Contains(">") || expression.Contains("<"));
        }

        // [MS-FSA] 2.1.4.4
        // The FileName is string compared with Expression using the following wildcard rules:
        // * (asterisk) Matches zero or more characters.
        // ? (question mark) Matches a single character.
        // DOS_DOT (" quotation mark) Matches either a period or zero characters beyond the name string.
        // DOS_QM (> greater than) Matches any single character or, upon encountering a period or end of name string, advances the expression to the end of the set of contiguous DOS_QMs.
        // DOS_STAR (< less than) Matches zero or more characters until encountering and matching the final . in the name.
        private static bool IsFileNameInExpression(string fileName, string expression)
        {
            if (expression == "*")
            {
                return true;
            }
            else if (expression.EndsWith("*")) // expression.Length > 1
            {
                var desiredFileNameStart = expression.Substring(0, expression.Length - 1);
                var findExactNameWithoutExtension = false;
                if (desiredFileNameStart.EndsWith("\""))
                {
                    findExactNameWithoutExtension = true;
                    desiredFileNameStart = desiredFileNameStart.Substring(0, desiredFileNameStart.Length - 1);
                }

                if (!findExactNameWithoutExtension)
                {
                    if (fileName.StartsWith(desiredFileNameStart, StringComparison.OrdinalIgnoreCase))
                    {
                        return true;
                    }
                }
                else
                {
                    if (fileName.StartsWith(desiredFileNameStart + ".", StringComparison.OrdinalIgnoreCase) ||
                        fileName.Equals(desiredFileNameStart, StringComparison.OrdinalIgnoreCase))
                    {
                        return true;
                    }
                }
            }
            else if (expression.StartsWith("<"))
            {
                var desiredFileNameEnd = expression.Substring(1);
                if (fileName.EndsWith(desiredFileNameEnd, StringComparison.OrdinalIgnoreCase))
                {
                    return true;
                }
            }
            else if (string.Equals(fileName, expression, StringComparison.OrdinalIgnoreCase))
            {
                return true;
            }
            return false;
        }

        private static List<QueryDirectoryFileInformation> FromFileSystemEntries(List<IFileSystemEntry> entries, FileInformationClass informationClass)
        {
            var result = new List<QueryDirectoryFileInformation>();
            foreach (var entry in entries)
            {
                var information = FromIFileSystemEntry(entry, informationClass);
                result.Add(information);
            }
            return result;
        }

        private static QueryDirectoryFileInformation FromIFileSystemEntry(IFileSystemEntry entry, FileInformationClass informationClass)
        {
            switch (informationClass)
            {
                case FileInformationClass.FileBothDirectoryInformation:
                {
                    var result = new FileBothDirectoryInformation();
                    result.CreationTime = entry.CreationTime;
                    result.LastAccessTime = entry.LastAccessTime;
                    result.LastWriteTime = entry.LastWriteTime;
                    result.ChangeTime = entry.LastWriteTime;
                    result.EndOfFile = (long)entry.Size;
                    result.AllocationSize = (long)GetAllocationSize(entry.Size);
                    result.FileAttributes = GetFileAttributes(entry);
                    result.EaSize = 0;
                    result.ShortName = GetShortName(entry.Name);
                    result.FileName = entry.Name;
                    return result;
                }
                case FileInformationClass.FileDirectoryInformation:
                {
                    var result = new FileDirectoryInformation();
                    result.CreationTime = entry.CreationTime;
                    result.LastAccessTime = entry.LastAccessTime;
                    result.LastWriteTime = entry.LastWriteTime;
                    result.ChangeTime = entry.LastWriteTime;
                    result.EndOfFile = (long)entry.Size;
                    result.AllocationSize = (long)GetAllocationSize(entry.Size);
                    result.FileAttributes = GetFileAttributes(entry);
                    result.FileName = entry.Name;
                    return result;
                }
                case FileInformationClass.FileFullDirectoryInformation:
                {
                    var result = new FileFullDirectoryInformation();
                    result.CreationTime = entry.CreationTime;
                    result.LastAccessTime = entry.LastAccessTime;
                    result.LastWriteTime = entry.LastWriteTime;
                    result.ChangeTime = entry.LastWriteTime;
                    result.EndOfFile = (long)entry.Size;
                    result.AllocationSize = (long)GetAllocationSize(entry.Size);
                    result.FileAttributes = GetFileAttributes(entry);
                    result.EaSize = 0;
                    result.FileName = entry.Name;
                    return result;
                }
                case FileInformationClass.FileIdBothDirectoryInformation:
                {
                    var result = new FileIdBothDirectoryInformation();
                    result.CreationTime = entry.CreationTime;
                    result.LastAccessTime = entry.LastAccessTime;
                    result.LastWriteTime = entry.LastWriteTime;
                    result.ChangeTime = entry.LastWriteTime;
                    result.EndOfFile = (long)entry.Size;
                    result.AllocationSize = (long)GetAllocationSize(entry.Size);
                    result.FileAttributes = GetFileAttributes(entry);
                    result.EaSize = 0;
                    result.ShortName = GetShortName(entry.Name);
                    result.FileId = 0;
                    result.FileName = entry.Name;
                    return result;
                }
                case FileInformationClass.FileIdFullDirectoryInformation:
                {
                    var result = new FileIdFullDirectoryInformation();
                    result.CreationTime = entry.CreationTime;
                    result.LastAccessTime = entry.LastAccessTime;
                    result.LastWriteTime = entry.LastWriteTime;
                    result.ChangeTime = entry.LastWriteTime;
                    result.EndOfFile = (long)entry.Size;
                    result.AllocationSize = (long)GetAllocationSize(entry.Size);
                    result.FileAttributes = GetFileAttributes(entry);
                    result.EaSize = 0;
                    result.FileId = 0;
                    result.FileName = entry.Name;
                    return result;
                }
                case FileInformationClass.FileNamesInformation:
                {
                    var result = new FileNamesInformation();
                    result.FileName = entry.Name;
                    return result;
                }
                default:
                {
                    throw new UnsupportedInformationLevelException();
                }
            }
        }

        public NTStatus GetFileSystemInformation(out FileSystemInformation result, FileSystemInformationClass informationClass)
        {
            switch (informationClass)
            {
                case FileSystemInformationClass.FileFsVolumeInformation:
                {
                    var information = new FileFsVolumeInformation();
                    information.SupportsObjects = false;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsSizeInformation:
                {
                    var information = new FileFsSizeInformation();
                    information.TotalAllocationUnits = m_fileSystem.Size / ClusterSize;
                    information.AvailableAllocationUnits = m_fileSystem.FreeSpace / ClusterSize;
                    information.SectorsPerAllocationUnit = ClusterSize / BytesPerSector;
                    information.BytesPerSector = BytesPerSector;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsDeviceInformation:
                {
                    var information = new FileFsDeviceInformation();
                    information.DeviceType = DeviceType.Disk;
                    information.Characteristics = DeviceCharacteristics.IsMounted;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsAttributeInformation:
                {
                    var information = new FileFsAttributeInformation();
                    information.FileSystemAttributes = FileSystemAttributes.UnicodeOnDisk;
                    information.MaximumComponentNameLength = 255;
                    information.FileSystemName = m_fileSystem.Name;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsControlInformation:
                {
                    var information = new FileFsControlInformation();
                    information.FileSystemControlFlags = FileSystemControlFlags.ContentIndexingDisabled;
                    information.DefaultQuotaThreshold = UInt64.MaxValue;
                    information.DefaultQuotaLimit = UInt64.MaxValue;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsFullSizeInformation:
                {
                    var information = new FileFsFullSizeInformation();
                    information.TotalAllocationUnits = m_fileSystem.Size / ClusterSize;
                    information.CallerAvailableAllocationUnits = m_fileSystem.FreeSpace / ClusterSize;
                    information.ActualAvailableAllocationUnits = m_fileSystem.FreeSpace / ClusterSize;
                    information.SectorsPerAllocationUnit = ClusterSize / BytesPerSector;
                    information.BytesPerSector = BytesPerSector;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                case FileSystemInformationClass.FileFsObjectIdInformation:
                {
                    result = null;
                    // STATUS_INVALID_PARAMETER is returned when the file system does not implement object IDs
                    // See: https://msdn.microsoft.com/en-us/library/cc232106.aspx
                    return NTStatus.STATUS_INVALID_PARAMETER;
                }
                case FileSystemInformationClass.FileFsSectorSizeInformation:
                {
                    var information = new FileFsSectorSizeInformation();
                    information.LogicalBytesPerSector = BytesPerSector;
                    information.PhysicalBytesPerSectorForAtomicity = BytesPerSector;
                    information.PhysicalBytesPerSectorForPerformance = BytesPerSector;
                    information.FileSystemEffectivePhysicalBytesPerSectorForAtomicity = BytesPerSector;
                    information.ByteOffsetForSectorAlignment = 0;
                    information.ByteOffsetForPartitionAlignment = 0;
                    result = information;
                    return NTStatus.STATUS_SUCCESS;
                }
                default:
                {
                    result = null;
                    return NTStatus.STATUS_INVALID_INFO_CLASS;
                }
            }
        }

        public NTStatus SetFileInformation(object handle, FileInformation information)
        {
            FileHandle fileHandle = (FileHandle)handle;
            if (information is FileBasicInformation)
            {
                FileBasicInformation basicInformation = (FileBasicInformation)information;
                bool isHidden = ((basicInformation.FileAttributes & FileAttributes.Hidden) > 0);
                bool isReadonly = (basicInformation.FileAttributes & FileAttributes.ReadOnly) > 0;
                bool isArchived = (basicInformation.FileAttributes & FileAttributes.Archive) > 0;
                try
                {
                    m_fileSystem.SetAttributes(fileHandle.Path, isHidden, isReadonly, isArchived);
                }
                catch (Exception ex)
                {
                    NTStatus status = ToNTStatus(ex);
                    Log(Severity.Verbose, "SetFileInformation: Failed to set file attributes on '{0}'. {1}.", fileHandle.Path, status);
                    return status;
                }

                try
                {
                    m_fileSystem.SetDates(fileHandle.Path, basicInformation.CreationTime, basicInformation.LastWriteTime, basicInformation.LastAccessTime);
                }
                catch (Exception ex)
                {
                    NTStatus status = ToNTStatus(ex);
                    Log(Severity.Verbose, "SetFileInformation: Failed to set file dates on '{0}'. {1}.", fileHandle.Path, status);
                    return status;
                }
                return NTStatus.STATUS_SUCCESS;
            }
            else if (information is FileRenameInformationType2)
            {
                FileRenameInformationType2 renameInformation = (FileRenameInformationType2)information;
                string newFileName = renameInformation.FileName;
                if (!newFileName.StartsWith(@"\"))
                {
                    newFileName = @"\" + newFileName;
                }

                if (fileHandle.Stream != null)
                {
                    fileHandle.Stream.Close();
                }

                // Note: it's possible that we just want to upcase / downcase a filename letter.
                try
                {
                    if (renameInformation.ReplaceIfExists && (m_fileSystem.GetEntry(newFileName) != null))
                    {
                        m_fileSystem.Delete(newFileName);
                    }
                    m_fileSystem.Move(fileHandle.Path, newFileName);
                    Log(Severity.Information, "SetFileInformation: Renamed '{0}' to '{1}'", fileHandle.Path, newFileName);
                }
                catch (Exception ex)
                {
                    NTStatus status = ToNTStatus(ex);
                    Log(Severity.Verbose, "SetFileInformation: Cannot rename '{0}' to '{1}'. {2}.", fileHandle.Path, newFileName, status);
                    return status;
                }
                fileHandle.Path = newFileName;
                return NTStatus.STATUS_SUCCESS;
            }
            else if (information is FileDispositionInformation)
            {
                if (((FileDispositionInformation)information).DeletePending)
                {
                    // We're supposed to delete the file on close, but it's too late to report errors at this late stage
                    if (fileHandle.Stream != null)
                    {
                        fileHandle.Stream.Close();
                    }

                    try
                    {
                        m_fileSystem.Delete(fileHandle.Path);
                        Log(Severity.Information, "SetFileInformation: Deleted '{0}'", fileHandle.Path);
                    }
                    catch (Exception ex)
                    {
                        NTStatus status = ToNTStatus(ex);
                        Log(Severity.Information, "SetFileInformation: Error deleting '{0}'. {1}.", fileHandle.Path, status);
                        return status;
                    }
                }
                return NTStatus.STATUS_SUCCESS;
            }
            else if (information is FileAllocationInformation)
            {
                long allocationSize = ((FileAllocationInformation)information).AllocationSize;
                try
                {
                    fileHandle.Stream.SetLength(allocationSize);
                }
                catch (Exception ex)
                {
                    NTStatus status = ToNTStatus(ex);
                    Log(Severity.Verbose, "SetFileInformation: Cannot set allocation for '{0}'. {1}.", fileHandle.Path, status);
                    return status;
                }
                return NTStatus.STATUS_SUCCESS;
            }
            else if (information is FileEndOfFileInformation)
            {
                long endOfFile = ((FileEndOfFileInformation)information).EndOfFile;
                try
                {
                    fileHandle.Stream.SetLength(endOfFile);
                }
                catch (Exception ex)
                {
                    NTStatus status = ToNTStatus(ex);
                    Log(Severity.Verbose, "SetFileInformation: Cannot set end of file for '{0}'. {1}.", fileHandle.Path, status);
                    return status;
                }
                return NTStatus.STATUS_SUCCESS;
            }
            else
            {
                return NTStatus.STATUS_NOT_IMPLEMENTED;
            }
        }
    }
}
